;
; Purpose:
;	To make a simple node that implements S.N.A.P in AVR assembler
;	and handles SCAN Command (DB1=0x00) when  packet sent to address 0
;	(broadcast).
;
;	*** Notice ***
;	SCAN Command is used when searching for nodes in large networks and
;	is not part of S.N.A.P. I like it anyway!
;
; Author:
;	Gran Nordahl, e99_gno@e.kth.se
;
; Credits:
;	High Tech Horizon, www.hth.com [S.N.A.P]
;	John C. Wren [CRC16]
;
; File, Version, Date:
;	http://www.e.kth.se/~e99_gno/snap/snap03.asm, 1.0, 2001-06-19
;
; Target:
;	AT90S8515/STK500 @ 3.69 MHz, but will probably work on any AVR
;	with an UART.
;
; Description:
;	*** Packet sent to ME (node's address) ***
;	DB1 of received packet is shown on the LEDs of STK500 (PORTB)
;	Data from switches (PINA) is sent in ACK packet as DB1
;
;	*** Broadcast, i.e. packet sent to address 0 ***
;	Supports only SCAN command, i.e. DB1=0x00, and expects that DB2-7
;	contains LOW_LIMIT3-1 & HIGH_LIMIT3-1:
;	|SYNC|HDB2|HDB1|DAB|SAB|0x00|LOW_LIMIT3|LOW_LIMIT2|LOW_LIMIT1|
;	HIGH_LIMIT3|HIGH_LIMIT2|HIGH_LIMIT1|CRC2|CRC1|
;	ACK is only sent to sender if LOW_LIMIT <= ME <= HIGH_LIMIT and then
;	without any data bytes.	This makes it easy to scan large networks for nodes!
;
;	*** Example: How to scan large networks with nodes supporting SCAN Command ***
;	1: Initial limits: LOW_LIMIT = 0x000000, HIGH_LIMIT = 0xffffff
;	2: Broadcast with DB1=0x00 (SCAN Command) and limits in DB2-7
;	3: Any response?
;		* Yes, but error in CRC or lots of data but no SYNC:
;			More than one node within limits ==> collisions ==>
;			narrow search and repeat from #2 until only one node
;			answers (binary search)
;		* Yes, got packet with correct CRC!:
;			Only one node within limits. Save address and quit
;			search or search with different limits (repeat from #2)
;			Remember to not include found nodes within search limits!
;			It's waste of time finding nodes more than once!
;		* No, timeout:
;			No node within limits ==> quit search or search with
;			other limits (repeat from #2)
;
;
; Accepted packet structure:
;	HDB2: 1-3 bytes destination/source adress, No flags, ACK request
;	HDB1: Command mode not supported, CRC16, 1 or 7 data bytes
;
; Support:
;	Ask kindly
;
; Terms of use:
;	For non commercial purposes you may use this code as you want
;
;
; For more information about S.N.A.P, go to http://www.hth.com/snap/
;



.include "8515def.inc"



.def	LOW_LIMIT3  = r8	;Only used if BROADCAST and SCAN Command (DB1 =0x00)
.def	LOW_LIMIT2  = r9	;*
.def	LOW_LIMIT1  = r10	;*
.def	HIGH_LIMIT3 = r11	;*
.def	HIGH_LIMIT2 = r12	;*
.def	HIGH_LIMIT1 = r13	;*
.def	CRC_H	= r14		;Current CRC value, 8 upper bits
.def	CRC_L	= r15		;Current CRC value, 8 lower bits
.def	TEMP	= r16		;General purpose register
.def	TEMP2	= r17		;General purpose register
.def	COUNT	= r18		;Used in sub CRC16
.def	DATA	= r19		;Stores data from the UART and is used in sub CRC16
.def	HDB2	= r20		;Stores HDB2 of incoming packet
.def	HDB1	= r21		;Stores HDB1 of incoming packet
.def	SAB3	= r22		;Stores SAB3 of incoming packet. Used later as DAB3.
.def	SAB2	= r23		;Stores SAB2 of incoming packet. Used later as DAB2.
.def	SAB1	= r24		;Stores SAB1 of incoming packet. Used later as DAB1.
.def	DB1	= r25		;Stores DB1  of incoming packet.
.def	BROADCAST=r26		;If packet sent to address 0 then BROADCAST != 0



.set	ME	= 0x010203	;Nodes adress. Notice: set mask below according to number of
				;bytes in ME. SS=00XX0000, XX=01/10/11
.set	SS	= 0b00110000	;OR mask for HDB2 when BROADCAST (sets SS)
.set	SYNC	= 0b01010100	;Syncronisation byte
.set	RX_HDB2	= 0b11110001	;1-3 bytes destination/source adress, No flags, ACK request.
.set	RX_HDB1	= 0b01000111	;Command mode not supported, CRC16, up to 7 data bytes.






;Init STACK
	ldi	TEMP, low(RAMEND)
	out	SPL, TEMP
	ldi	TEMP, high(RAMEND)
	out	SPH, TEMP

;Init UART
	ldi	TEMP, 95	;95 -> 2400bps, 47 -> 4800bps,
	out	UBRR, TEMP	; 23 -> 9600bps, 11 -> 19200bps @ 3.69MHz
	ldi	TEMP, 0b00011000;Receiver and Transmitter enabled
	out	UCR, TEMP
	
;Init PORTA
	clr	TEMP
	out	DDRA, TEMP	;PORTA = input
		
;Init PORTB
	ser	TEMP
	out	DDRB, TEMP	;PORTB = output
	out	PORTB, TEMP	;All LEDs cleared (LEDs inverted on STK500)
	


NO_SYNC:
	clr	CRC_H		;Reset CRC16 
	clr	CRC_L
	
	clr	HIGH_LIMIT3	;Resets address limits
	clr	HIGH_LIMIT2
	clr	HIGH_LIMIT1
	clr	LOW_LIMIT3
	clr	LOW_LIMIT2
	clr	LOW_LIMIT1	
	
	clr	BROADCAST	;No broadcast (default)
	
	rcall	RX_LOOP		;Wait for new byte (SYNC)
	cpi	DATA, SYNC
	brne	NO_SYNC		;Received byte was not SYNC
	
	rcall	RX_LOOP		;Wait for new byte (HDB2)
	mov	HDB2, DATA	;Store HDB2
	sbr	DATA, 0b11110000;This allows 1-3 DAB & SAB
	cpi	DATA, RX_HDB2	
	brne	NO_SYNC		;HDB2 not OK for me ==> Skip packet
	mov	DATA, HDB2	;Sub CRC16 always called with data in DATA
	rcall	CRC16		;Calculate CRC16
	
	rcall	RX_LOOP		;Wait for new byte (HDB1)
	mov	HDB1, DATA	;Store HDB1
	sbr	DATA, 0b00000111;This allows up to 7 data bytes (but only 1 or 7 used)
	cpi	DATA, RX_HDB1
	brne	NO_SYNC		;HDB1 not OK for me ==> Skip packet
	mov	DATA, HDB1	;Sub CRC16 always called with data in DATA
	rcall	CRC16		;Calculate CRC16



;1-3 DAB?:
	mov	TEMP, HDB2
	cbr	TEMP, 0b00111111;Mask out DD
	cpi	TEMP, 0b11000000;3 DAB? (DD == 11)
	breq	R_DAB3		;Yes
	cpi	TEMP, 0b10000000;No: 2 DAB? (DD == 10)
	breq	R_DAB2		;Yes
	cpi	TEMP, 0b01000000;No: 1 DAB? (DD == 01)
	breq	R_DAB1		;Yes

	;If this happens we got an evil error!
	rjmp	NO_SYNC		;0 DAB ==> Skip packet	


	
R_DAB3:	rcall	RX_LOOP		;Wait for new byte (DAB3)
	cpi	DATA, byte3(ME)
	brne	NO_SYNC		;DAB3 not me ==> Skip packet
	rcall	CRC16		;Calculate CRC16

R_DAB2:	rcall	RX_LOOP		;Wait for new byte (DAB2)
	cpi	DATA, high(ME)
	brne	NO_SYNC		;DAB2 not me ==> Skip packet
	rcall	CRC16		;Calculate CRC16

R_DAB1:	rcall	RX_LOOP		;Wait for new byte (DAB1)
	mov	TEMP2, DATA	;DATA needed after "rcall CRC16" ==> stored in TEMP2
	rcall	CRC16		;Calculate CRC16
	cpi	TEMP2, low(ME)	;DAB1 me?
	brne	R_BROADCAST	;No: Broadcast? (DAB1 == 0 && DD == 01)
	rjmp	R_SAB		;Yes



R_BROADCAST:
	mov	TEMP, HDB2
	cbr	TEMP, 0b00111111;Mask out DD
	cpi	TEMP, 0b01000000;1 DAB? (DD == 01)
	breq	R_BROADCAST2
	rjmp	NO_SYNC		;DD != 01 ==> Skip packet

R_BROADCAST2:
	cpi	TEMP2, 0x00
	breq	R_BROADCAST3
	rjmp	NO_SYNC		;DAB1 != 0 ==> Skip packet	

R_BROADCAST3:
	ldi	BROADCAST, 0x01	;BROADCAST!=0 indicates that packet was sent to address 0

	

R_SAB:
;1-3 SAB?:
	mov	TEMP, HDB2
	cbr	TEMP, 0b11001111;Mask out SS
	cpi	TEMP, 0b00110000;3 SAB?
	breq	R_SAB3		;Yes
	cpi	TEMP, 0b00100000;No: 2 SAB?
	breq	R_SAB2		;Yes
	cpi	TEMP, 0b00010000;No: 1 SAB?
	breq	R_SAB1		;Yes

	;If this happens we got an evil error!
	rjmp	NO_SYNC		;0 DAB ==> Skip packet



R_SAB3:	rcall	RX_LOOP		;Wait for new byte (SAB3)
	mov	SAB3, DATA	;Save SAB3
	rcall	CRC16		;Calculate CRC16

R_SAB2:	rcall	RX_LOOP		;Wait for new byte (SAB2)
	mov	SAB2, DATA	;Save SAB2
	rcall	CRC16		;Calculate CRC16

R_SAB1:	rcall	RX_LOOP		;Wait for new byte (SAB1)
	mov	SAB1, DATA	;Save SAB1
	rcall	CRC16		;Calculate CRC16

	rcall	RX_LOOP		;Wait for new byte (DB1)
	mov	DB1, DATA	;Save DB1
	rcall	CRC16		;Calculate CRC16
	tst	BROADCAST	;Broadcast?
	breq	R_CRC2		;No: No more data ==> Calculate CRC2

	cpi	DB1, 0x00	;Yes: SCAN Command? (DB1=0x00)
	breq	R_DB		;	Yes
	rjmp	NO_SYNC		;	No: Not supported command ==> Skip packet

R_DB:	mov	TEMP, HDB1
	cbr	TEMP, 0b11110000
	cpi	TEMP, 0b00000111;NNNN == 0111? (7 bytes)
	breq	R_DB2		;Yes
	rjmp	NO_SYNC		;No: NNNN != 0111 ==> Skip packet

R_DB2:	rcall	RX_LOOP		;Wait for new byte (DB2)
	mov	LOW_LIMIT3, DATA;Save LOW_LIMIT3
	rcall	CRC16		;Calculate CRC16
R_DB3:	rcall	RX_LOOP		;Wait for new byte (DB3)
	mov	LOW_LIMIT2, DATA;Save LOW_LIMIT2
	rcall	CRC16		;Calculate CRC16
R_DB4:	rcall	RX_LOOP		;Wait for new byte (DB4)
	mov	LOW_LIMIT1, DATA;Save LOW_LIMIT1
	rcall	CRC16		;Calculate CRC16
R_DB5:	rcall	RX_LOOP		;Wait for new byte (DB5)
	mov	HIGH_LIMIT3, DATA;Save HIGH_LIMIT3
	rcall	CRC16		;Calculate CRC16
R_DB6:	rcall	RX_LOOP		;Wait for new byte (DB6)
	mov	HIGH_LIMIT2, DATA;Save HIGH_LIMIT2
	rcall	CRC16		;Calculate CRC16
R_DB7:	rcall	RX_LOOP		;Wait for new byte (DB7)
	mov	HIGH_LIMIT1, DATA;Save HIGH_LIMIT1
	rcall	CRC16		;Calculate CRC16	



R_CRC2:
	rcall	RX_LOOP		;Wait for new byte (CRC2)
	rcall	CRC16		;Calculate CRC16
	rcall	RX_LOOP		;Wait for new byte (CRC1)
	rcall	CRC16		;Calculate CRC16
	

	
;Prepares HDB2 & HDB1:
	mov	TEMP, HDB2		;The following swaps DD and SS of HDB2
	cbr	HDB2, 0b00111111
	lsr	HDB2
	lsr	HDB2
	cbr	TEMP, 0b11001111	
	lsl	TEMP
	lsl	TEMP	
	add	HDB2, TEMP		;DD and SS of HDB2 swapped!

	
	;Broadcast ==> DD always 01. ME not always 1 byte ==> problem, which is solwed by
	;the following four lines:  
	tst	BROADCAST		;Broadcast?
	breq	T_ACKR			;No
	cbr	HDB2, 0b00110000	;Yes: SS cleared
	sbr	HDB2, SS		;     SS set according to the number of bytes in ME



T_ACKR:	sbr	HDB2, 0b00000010	;ACK response
	tst	CRC_H
	brne	T_NAK			;If CRC16 != 0, then create a NAK-packet
	tst	CRC_L
	breq	T_ACK			;If CRC16 != 0, then create a NAK-packet
T_NAK:	sbr	HDB2, 0b00000001	;NAK response
	ldi	HDB1, 0b01000000	;Command mode not supported, CRC16, 0 Data bytes
	clr	CRC_H
	clr	CRC_L
	rjmp	T_SYNC

T_ACK:	ldi	HDB1, 0b01000001	;Command mode not supported, CRC16, 1 Data byte

	tst	BROADCAST		;Broadcast?
	brne	T_LOW3			;Yes



; Shows DB1 on LEDs if not broadcast
	com	DB1
	out	PORTB, DB1	;Inverted DB1 to LEDs (LEDs inverted on STK500)
	rjmp	T_SYNC

	

T_LOW3:	ldi	HDB1, 0b01000000	;Command mode not supported, CRC16, 0 Data bytes
	mov	TEMP, LOW_LIMIT3	
	cpi	TEMP, byte3(ME)		;LOW_LIMIT3 <= byte3(ME)?
	breq	T_LOW2			;Yes [==]: Compare with LOW_LIMIT2
	brlo	T_HIGH3			;Yes [<]:  Skip LOW_LIMIT2-1, compare with HIGH_LIMIT3   
	rjmp	NO_SYNC			;No  [>]:  Out of bounds ==> Skip packet
T_LOW2:	mov	TEMP, LOW_LIMIT2
	cpi	TEMP, high(ME)		;LOW_LIMIT2 <= high(ME)?
	breq	T_LOW1			;Yes [==]: Compare with LOW_LIMIT1
	brlo	T_HIGH3			;Yes [<]:  Skip LOW_LIMIT1, compare with HIGH_LIMIT3
	rjmp	NO_SYNC			;No  [>]:  Out of bounds ==> Skip packet
T_LOW1:	mov	TEMP, LOW_LIMIT1
	cpi	TEMP, low(ME)		;LOW_LIMIT1 <= low(ME)?
	breq	T_HIGH3			;Yes [==]: Compare with HIGH_LIMIT3
	brlo	T_HIGH3			;Yes [<]:  Compare with HIGH_LIMIT3	
	rjmp	NO_SYNC			;No  [>]:  Out of bounds ==> Skip packet
T_HIGH3:mov	TEMP, HIGH_LIMIT3
	cpi	TEMP, byte3(ME)		;HIGH_LIMIT3 >= byte3(ME)?
	breq	T_HIGH2			;Yes [==]: Compare with HIGH_LIMIT2
	brsh	T_SYNC			;Yes [>=]: Send ACK packet
	rjmp	NO_SYNC			;No  [<]:  Out of bounds ==> Skip packet
T_HIGH2:mov	TEMP, HIGH_LIMIT2
	cpi	TEMP, high(ME)		;HIGH_LIMIT2 >= high(ME)?
	breq	T_HIGH1			;Yes [==]: Compare with HIGH_LIMIT1
	brsh	T_SYNC			;Yes [>=]: Send ACK packet
	rjmp	NO_SYNC			;No  [<]:  Out of bounds ==> Skip packet
T_HIGH1:mov	TEMP, HIGH_LIMIT1
	cpi	TEMP, low(ME)		;HIGH_LIMIT1 >= low(ME)?
	breq	T_SYNC			;Yes [==]: Send ACK packet
	brsh	T_SYNC			;Yes [>=]: Send ACK packet	
	rjmp	NO_SYNC			;No  [<]:  Out of bounds ==> Skip packet



T_SYNC:	ldi	DATA, SYNC	;Transmit SYNC
	rcall	TX
	
	mov	DATA, HDB2	;Transmit HDB2
	rcall	CRC16		;Calculate CRC16
	rcall	TX
	
	mov	DATA, HDB1	;Transmit HDB1
	rcall	CRC16		;Calculate CRC16
	rcall	TX



;1-3 DAB?:
	mov	TEMP, HDB2
	cbr	TEMP, 0b00111111;Mask out DD
	cpi	TEMP, 0b11000000;3 DAB?
	breq	T_DAB3		;Yes
	cpi	TEMP, 0b10000000;No: 2 DAB?
	breq	T_DAB2		;Yes
	cpi	TEMP, 0b01000000;No: 1 DAB?
	breq	T_DAB1		;Yes
	
	;If this happens we got an evil error!
	rjmp	NO_SYNC		;0 DAB ==> Skip packet


	
T_DAB3:	mov	DATA, SAB3	;SAB3 of received packet now acts as DAB3
	rcall	CRC16		;Calculate CRC16
	rcall	TX

T_DAB2:	mov	DATA, SAB2	;SAB2 of received packet now acts as DAB2 
	rcall	CRC16		;Calculate CRC16
	rcall	TX

T_DAB1:	mov	DATA, SAB1	;SAB1 of received packet now acts as DAB1 
	rcall	CRC16		;Calculate CRC16
	rcall	TX



;1-3 SAB?:
	mov	TEMP, HDB2
	cbr	TEMP, 0b11001111;Mask out SS
	cpi	TEMP, 0b00110000;3 SAB?
	breq	T_SAB3		;Yes
	cpi	TEMP, 0b00100000;2 SAB?
	breq	T_SAB2		;Yes
	cpi	TEMP, 0b00010000;1 SAB?
	breq	T_SAB1		;Yes

	;If this happens we got an evil error!
	rjmp	NO_SYNC		;0 DAB ==> Skip packet



T_SAB3:	ldi	DATA, byte3(ME)	;Sends bit 23-16 of ME as SAB3
	rcall	CRC16		;Calculate CRC16
	rcall	TX

T_SAB2:	ldi	DATA, high(ME)	;Sends bit 15-8 of ME as SAB2
	rcall	CRC16		;Calculate CRC16
	rcall	TX
	
T_SAB1:	ldi	DATA, low(ME)	;Sends bit 7-0 of ME as SAB1
	rcall	CRC16		;Calculate CRC16
	rcall	TX
	cbr	HDB1, 0b11110000;1 or 0 Data Bytes?
	breq	T_DB0

T_DB1:	in	DATA, PINA	;PINA in DB1
	com	DATA		;Switches inverted on STK500
	rcall	CRC16		;Calculate CRC16
	rcall	TX
	
T_DB0:	mov	DATA, CRC_H	;Sends CRC2
	rcall	TX

	mov	DATA, CRC_L	;Sends CRC1
	rcall	TX

	rjmp	NO_SYNC		;Done! Let's do it again!
	
	




;*** RX_LOOP ***
;Returns received byte in DATA
RX_LOOP:
	sbis	USR, RXC
	rjmp	RX_LOOP
	in	DATA, UDR
	ret

	
	



;*** TX ***
;Sends DATA
TX:	sbi	USR, TXC
	out	UDR, DATA
TX_LOOP:
	sbis	USR, TXC
	rjmp	TX_LOOP
	ret






;*** CRC16 ***
;Purpose:
;	Calculate CRC-16 Values. Uses The CCITT-16 Polynomial,
;	Expressed As X^16 + X^12 + X^5 + 1
;Author:
;	John C. Wren
;Origin:
;	http://www.kd4dts.com/download/zip/crc16.zip  [Crc16.asm]
;2000-07-17:
;	Modifyed from MCS-51 to AVR by Gran Nordahl

CRC16:	eor	CRC_H, DATA
	ldi	COUNT, 0x08
CRC16_LOOP:
	lsl	CRC_L
	rol	CRC_H
	brcc	CRC16_SKIP
	ldi	TEMP, 0x10
	eor	CRC_H, TEMP
	ldi	TEMP, 0x21
	eor	CRC_L, TEMP
CRC16_SKIP:
	dec	COUNT
	brne	CRC16_LOOP
	ret
;Vulpes^2